# coding: utf8
from __future__ import unicode_literals

from .neural._classes.batchnorm import BatchNorm  # noqa: F401
from .neural._classes.layernorm import LayerNorm  # noqa: F401
from .neural._classes.resnet import Residual  # noqa: F401
from .neural._classes.feature_extracter import FeatureExtracter  # noqa: F401
from .neural._classes.function_layer import FunctionLayer  # noqa: F401
from .neural._classes.feed_forward import FeedForward  # noqa: F401
from ._registry import registry


@registry.layers.register("FeatureExtractor.v1")
def make_FeatureExtractor(attrs):
    return FeatureExtracter(attrs)


@registry.layers.register("LayerNorm.v1")
def make_LayerNorm(outputs=None, child=None):
    return LayerNorm(nO=outputs, child=child)



# coding: utf8
from __future__ import unicode_literals

from ..neural._classes.affine import Affine


def get_model(W_b_input, cls=Affine):
    W, b, input_ = W_b_input
    nr_out, nr_in = W.shape
    model = cls(nr_out, nr_in)
    model.W[:] = W
    model.b[:] = b
    return model


def get_shape(W_b_input):
    W, b, input_ = W_b_input
    return input_.shape[0], W.shape[0], W.shape[1]



# coding: utf8
from __future__ import unicode_literals

import numpy

from ...linear.linear import LinearModel


def test_init():
    model = LinearModel(3)
    keys = numpy.ones((5,), dtype="uint64")
    values = numpy.ones((5,), dtype="f")
    lengths = numpy.zeros((2,), dtype=numpy.int_)
    lengths[0] = 3
    lengths[1] = 2
    scores, backprop = model.begin_update((keys, values, lengths))
    assert scores.shape == (2, 3)
    d_feats = backprop(scores)
    assert d_feats is None



# coding: utf8
from __future__ import unicode_literals

from .. import Model
from ...api import layerize
from ... import describe
from ...describe import Dimension, Synapses, Gradient


def inverse(total):
    inverse = 1.0 / (1 + total)

    def backward(d_inverse):
        result = d_inverse * (-1 / (total + 1) ** 2)
        return result

    return inverse, backward


def _get_mask(ops, shape, drop):
    return ops.xp.random.uniform(0.0, 1.0, shape) > drop


def Siamese(layer, similarity):
    def begin_update(inputs, drop=0.0):
        ops = layer.ops
        if drop not in (None, 0.0):
            dropped = []
            for in1, in2 in inputs:
                if in1.size > in2.size:
                    mask = _get_mask(ops, in1.shape, drop)
                else:
                    mask = _get_mask(ops, in2.shape, drop)
                in1 = in1 * mask[: in1.shape[0]]
                in2 = in2 * mask[: in2.shape[0]]
                dropped.append((in1, in2))
            inputs = dropped

        input1, input2 = zip(*inputs)
        vec1, bp_vec1 = layer.begin_update(input1, drop=0.0)
        vec2, bp_vec2 = layer.begin_update(input2, drop=0.0)
        output, bp_output = similarity.begin_update((vec1, vec2), drop=0.0)

        def finish_update(d_output, sgd=None):
            d_vec1, d_vec2 = bp_output(d_output, sgd)
            # Remember that this is the same layer --
            # Is this bad? Are we making bp_vec2 stale?
            d_input1 = bp_vec1(d_vec1, lambda *args, **kwargs: None)
            d_input2 = bp_vec2(d_vec2, sgd)
            return (d_input1, d_input2)

        return output, finish_update

    model = layerize(begin_update)

    model._layers.append(layer)
    model._layers.append(similarity)

    def on_data(self, X, y):
        input1, input2 = zip(*X)
        for hook in layer.on_data_hooks:
            hook(layer, input1, y)

    model.on_data_hooks.append(on_data)
    return model


def unit_init(W, ops):
    W.fill(1)


@describe.attributes(
    nO=Dimension("Output size"),
    W=Synapses("Weights matrix", lambda obj: (obj.nO,), unit_init),
    d_W=Gradient("W"),
)
class CauchySimilarity(Model):
    # From chen (2013)
    def __init__(self, length):
        Model.__init__(self)
        self.nO = length

    def begin_update(self, vec1_vec2, drop=0.0):
        weights = self.W
        vec1, vec2 = vec1_vec2
        diff = vec1 - vec2
        square_diff = diff ** 2
        total = (weights * square_diff).sum(axis=1)
        sim, bp_sim = inverse(total)
        total = total.reshape((vec1.shape[0], 1))

        def finish_update(d_sim, sgd=None):
            d_total = bp_sim(d_sim)
            d_total = d_total.reshape(total.shape)
            self.d_W += (d_total * square_diff).sum(axis=0)
            d_square_diff = weights * d_total
            d_diff = 2 * d_square_diff * diff
            d_vec1 = d_diff
            d_vec2 = -d_diff
            sgd(self._mem.weights, self._mem.gradient, key=self.id)
            return (d_vec1, d_vec2)

        return sim, finish_update



# coding: utf8
from __future__ import unicode_literals

from .neural.pooling import Pooling, max_pool, mean_pool, sum_pool  # noqa: F401
from ._registry import registry


@registry.layers.register("MeanPooling.v1")
def make_MeanPooling_v1():
    return Pooling(mean_pool)


@registry.layers.register("SumPooling.v1")
def make_SumPooling_v1():
    return Pooling(sum_pool)


@registry.layers.register("MaxPooling.v1")
def make_MaxPooling_v1():
    return Pooling(max_pool)


@registry.layers.register("MeanMaxPooling.v1")
def make_MeanMaxPooling_v1():
    return Pooling(mean_pool, max_pool)



# coding: utf8
from __future__ import unicode_literals

import pytest
from srsly import cloudpickle as pickle

from ...api import with_flatten
from ...v2v import Affine


@pytest.fixture
def affine():
    return Affine(5, 3)


def test_pickle_with_flatten(affine):
    Xs = [affine.ops.allocate((2, 3)), affine.ops.allocate((4, 3))]
    model = with_flatten(affine)
    pickled = pickle.dumps(model)
    loaded = pickle.loads(pickled)
    Ys = loaded(Xs)
    assert len(Ys) == 2
    assert Ys[0].shape == (Xs[0].shape[0], affine.nO)
    assert Ys[1].shape == (Xs[1].shape[0], affine.nO)



# coding: utf8
from __future__ import unicode_literals

import numpy
import numpy.random
import itertools
import srsly
import tqdm


def minibatch(train_X, train_y, size=16, nr_update=1000):
    with tqdm.tqdm(total=nr_update * size, leave=False) as pbar:
        while nr_update >= 0:
            indices = numpy.arange(len(train_X))
            numpy.random.shuffle(indices)
            j = 0
            while j < indices.shape[0]:
                slice_ = indices[j : j + size]
                X = _take_slice(train_X, slice_)
                y = _take_slice(train_y, slice_)
                yield X, y
                j += size
                nr_update -= 1
                if nr_update <= 0:
                    break
                pbar.update(size)


def _take_slice(data, slice_):
    if isinstance(data, list) or isinstance(data, tuple):
        return [data[int(i)] for i in slice_]
    else:
        return data[slice_]


class BestFirstFinder(object):
    def __init__(self, **param_values):
        self.queue = []
        self.limit = 16
        self.params = param_values
        self.best_acc = 0.0
        self.best_i = 0
        self.i = 0
        self.j = 0
        self.best_model = None
        self.temperature = 0.0

    @property
    def configs(self):
        keys, value_groups = zip(*self.params.items())
        for values in itertools.product(*value_groups):
            config = dict(zip(keys, values))
            yield config

    def enqueue(self, model, train_acc, check_acc):
        fom = check_acc * min(check_acc / train_acc, 1.0)
        self.queue.append([fom, self.i, 0, model])
        if check_acc >= self.best_acc:
            self.best_acc = check_acc
            self.best_i = self.i
            self.best_model = model
            self.temperature = 0.0
        else:
            self.temperature += 0.01
        self.j = 0
        self.queue.sort(reverse=True)
        self.queue = self.queue[: self.limit]

    def __iter__(self):
        self.queue.sort(reverse=True)
        self.queue = self.queue[: self.limit]
        for i in range(len(self.queue)):
            self.queue[i][0] = self.queue[i][0] - 0.01
            self.queue[i][-1][2]["parent"] = self.queue[i][2]
            self.queue[i][2] += 1
            yield self.queue[i][-1]

    @property
    def best(self):
        return self.best_model


def resample_hyper_params(hparams, temperature):
    hparams = dict(hparams)
    hparams["epochs"] = hparams.get("epochs", 0) + 1
    hparams["learn_rate"] = resample(hparams["learn_rate"], 1e-6, 0.1, temperature)
    # hparams['beta1'] = resample(hparams.get('beta1', 0.9), 0.8, 1.0, temperature)
    # hparams['beta2'] = resample(hparams.get('beta2', 0.9), 0.8, 1.0, temperature)
    # hparams['L2'] = resample(hparams['L2'], 0.0, 1e-3, temperature)
    # hparams['batch_size'] = int(resample(hparams['batch_size'], 10, 256, temperature))
    # hparams['dropout'] = resample(hparams['dropout'], 0.05, 0.7, temperature)
    return hparams


def resample(curr, min_, max_, temperature):
    if temperature == 0.0:
        return curr
    scale = (max_ - min_) * temperature
    next_ = numpy.random.normal(loc=curr, scale=scale)
    return min(max_, max(min_, next_))


def train_epoch(
    model, sgd, hparams, train_X, train_y, dev_X, dev_y, device_id=-1, temperature=0.0
):
    model, sgd, hparams = srsly.pickle_loads(srsly.pickle_dumps((model, sgd, hparams)))
    if device_id >= 0:
        model.to_gpu(device_id)
        sgd.ops = model.ops
        sgd.to_gpu()
        if isinstance(train_y, numpy.ndarray):
            train_y = model.ops.asarray(train_y)
            dev_y = model.ops.asarray(dev_y)
    hparams = resample_hyper_params(hparams, temperature)
    sgd.learn_rate = hparams["learn_rate"]
    sgd.beta1 = hparams["beta1"]
    sgd.beta2 = hparams["beta2"]
    sgd.L2 = hparams["L2"]
    train_acc = 0.0
    train_n = 0
    for X, y in minibatch(
        train_X, train_y, size=hparams["batch_size"], nr_update=hparams["nr_update"]
    ):
        yh, finish_update = model.begin_update(X, drop=hparams["dropout"])
        if hasattr(y, "shape"):
            dy = (yh - y) / y.shape[0]
            train_acc += (y.argmax(axis=1) == yh.argmax(axis=1)).sum()
            train_n += y.shape[0]
        else:
            n_y = sum(len(y_i) for y_i in y)
            dy = [(yh[i] - y[i]) / n_y for i in range(len(yh))]
            for i in range(len(y)):
                train_acc += (y[i].argmax(axis=1) == yh[i].argmax(axis=1)).sum()
            train_n += n_y
        finish_update(dy, sgd=sgd)
    train_acc /= train_n
    with model.use_params(sgd.averages):
        dev_acc = model.evaluate(dev_X, dev_y)
    model.to_cpu()
    sgd.to_cpu()
    return device_id, ((model, sgd, hparams), float(train_acc), float(dev_acc))


class DevicePool(object):
    """Synchronize GPU usage"""

    def __init__(self, n):
        self.devices = {i: None for i in range(n)}

    def acquire(self):
        for i, device in self.devices.items():
            if device is None:
                self.devices[i] = True
                return i
        else:
            return None

    def release(self, i):
        if i in self.devices:
            self.devices[i] = None


#
# def best_first_sgd(initials, train_X, train_y, dev_X, dev_y,
#        get_new_model=None, get_score=None):
#    if get_new_model is None:
#        get_new_model = _get_new_model
#    if get_score is None:
#        get_score = _get_score
#
#    queue = []
#    for i, model in enumerate(initials):
#        train_acc, model = get_new_model(model, train_X, train_y)
#        check_acc = get_score(model, dev_X, dev_y)
#        ratio = min(check_acc / train_acc, 1.0)
#        print((model[-1], train_acc, check_acc))
#        queue.append([check_acc * ratio, i, model])
#
#    train_acc = 0
#    limit = 8
#    i = 0
#    best_model = None
#    best_acc = 0.0
#    best_i = 0
#    while best_i > (i - 100) and train_acc < 0.999:
#        queue.sort(reverse=True)
#        queue = queue[:limit]
#        prev_score, parent, model = queue[0]
#        queue[0][0] -= 0.001
#        yield prev_score, parent, model
#        train_acc, new_model = get_new_model(model, train_X, train_y)
#        check_acc = get_score(new_model, dev_X, dev_y)
#        ratio = min(check_acc / train_acc, 1.0)
#
#        i += 1
#        queue.append([check_acc * ratio, i, new_model])
#
#        if check_acc >= best_acc:
#            best_acc = check_acc
#            best_i = i
#            best_model = new_model
#        progress = {
#            'i': i,
#            'parent': parent,
#            'prev_score': prev_score,
#            'this_score': queue[-1][0],
#            'train_acc': train_acc,
#            'check_acc': check_acc,
#            'best_acc': best_acc,
#            'hparams': new_model[-1]
#        }
#        yield best_model, progress
#
#
#



# encoding: utf8
from __future__ import unicode_literals

import tempfile
import os
import pytest
import threading
import time

from ...neural._classes import model as base
from ...neural.ops import NumpyOps


@pytest.fixture
def model_with_no_args():
    model = base.Model()
    return model


def test_Model_defaults_to_name_model(model_with_no_args):
    assert model_with_no_args.name == "model"


def test_changing_instance_name_doesnt_change_class_name():
    model = base.Model()
    assert model.name != "changed"
    model.name = "changed"
    model2 = base.Model()
    assert model2.name != "changed"


def test_changing_class_name_doesnt_change_default_instance_name():
    model = base.Model()
    assert model.name != "changed"
    base.Model.name = "changed"
    assert model.name != "changed"
    # Reset state
    base.Model.name = "model"


def test_changing_class_name_doesnt_changes_nondefault_instance_name():
    model = base.Model(name="nondefault")
    assert model.name == "nondefault"
    base.Model.name = "changed"
    assert model.name == "nondefault"


def test_Model_defaults_to_cpu(model_with_no_args):
    assert isinstance(model_with_no_args.ops, NumpyOps)


def test_models_get_different_ids(model_with_no_args):
    model1 = base.Model()
    model2 = base.Model()
    assert model1.id != model2.id


def test_init_assigns_attributes():
    model = base.Model()
    model._mem
    assert model._layers == []


def test_init_installs_via_descriptions():
    def mock_install(attr, self):
        setattr(self, attr, "model=" + self.name)

    base.Model.descriptions = [("myattr", mock_install)]
    model = base.Model(name="model1")
    assert model.myattr == "model=%s" % "model1"
    model2 = base.Model(name="model2")
    assert model2.myattr == "model=%s" % "model2"


def test_init_calls_hooks():
    def mock_init_hook(self, *args, **kwargs):
        setattr(self, "hooked", (args, kwargs))

    base.Model.on_init_hooks = [mock_init_hook]
    model = base.Model(0, 1, 2)
    assert model.hooked == ((0, 1, 2), {})
    model2 = base.Model(value="something")
    assert model2.hooked == (tuple(), {"value": "something"})


def test_use_device():
    dev_id = id(base.Model.ops)
    with base.Model.use_device(base.Model.ops.device):
        assert id(base.Model.ops) == dev_id
    with base.Model.use_device("gpu"):
        assert id(base.Model.ops) != dev_id
    assert id(base.Model.ops) == dev_id


def test_bind_plus():
    with base.Model.define_operators({"+": lambda a, b: (a.name, b.name)}):
        m = base.Model(name="a") + base.Model(name="b")
        assert m == ("a", "b")


def test_plus_chain():
    with base.Model.define_operators({"+": lambda a, b: a}):
        m = (
            base.Model(name="a")
            + base.Model(name="b")
            + base.Model(name="c")
            + base.Model(name="d")
        )
        assert m.name == "a"


def test_overload_operators_in_subthread():
    """Test we can create a model in a child thread with overloaded operators."""
    # Worker1 will start and run, while worker 2 sleeps after Model.define_operators.
    # Without thread-safety, worker2 will find that its operator definitions
    # have been removed, causing an error.
    worker1 = threading.Thread(target=_overload_plus, args=("+", 0))
    worker2 = threading.Thread(target=_overload_plus, args=("*", 1,))
    worker2.start()
    worker1.start()
    worker1.join()
    worker2.join()

    worker1 = threading.Thread(target=_overload_plus, args=("+", 1))
    worker2 = threading.Thread(target=_overload_plus, args=("*", 0,))
    worker2.start()
    worker1.start()
    worker1.join()
    worker2.join()


def _overload_plus(operator, sleep):
    m1 = base.Model(name="a")
    m2 = base.Model(name="b")
    with base.Model.define_operators({operator: lambda a, b: a.name + b.name}):
        time.sleep(sleep)
        if operator == "+":
            value = m1 + m2
        else:
            value = m1 * m2
    assert value == "ab"
    assert base.Model._thread_local.operators == {}


def test_nested_operator_contexts():
    operator = "+"
    m1 = base.Model(name="a")
    m2 = base.Model(name="b")
    assert base.Model._thread_local.operators == {}
    with base.Model.define_operators({"+": lambda a, b: a.name + b.name}):
        value = m1 + m2
        with pytest.raises(TypeError):
            value = m1 * m2
        with base.Model.define_operators({"*": lambda a, b: a.name + b.name}):
            with pytest.raises(TypeError):
                value = m1 + m2
            value = m1 * m2
            with base.Model.define_operators({"-": lambda a, b: a.name + b.name}):
                with pytest.raises(TypeError):
                    value = m1 + m2
                value = m1 - m2
            with pytest.raises(TypeError):
                value = m1 + m2
            value = m1 * m2
        value = m1 + m2
        with pytest.raises(TypeError):
            value = m1 * m2
    assert value == "ab"
    assert base.Model._thread_local.operators == {}


@pytest.mark.parametrize("op", "+ - * @ / // % ** << >> & ^ |".split())
def test_all_operators(op):
    m1 = base.Model(name="a")
    m2 = base.Model(name="b")
    with base.Model.define_operators({op: lambda a, b: a.name + b.name}):
        if op == "+":
            value = m1 + m2
        else:
            with pytest.raises(TypeError):
                value = m1 + m2
        if op == "-":
            value = m1 - m2
        else:
            with pytest.raises(TypeError):
                value = m1 - m2

        if op == "*":
            value = m1 * m2
        else:
            with pytest.raises(TypeError):
                value = m1 * m2

        if op == "@":
            value = m1.__matmul__(m2)  # Be kind to Python 2...
        else:
            with pytest.raises(TypeError):
                value = m1.__matmul__(m2)

        if op == "/":
            value = m1 / m2
        else:
            with pytest.raises(TypeError):
                value = m1 / m2

        if op == "//":
            value = m1 // m2
        else:
            with pytest.raises(TypeError):
                value = m1 // m2
        if op == "^":
            value = m1 ^ m2
        else:
            with pytest.raises(TypeError):
                value = m1 ^ m2
        if op == "%":
            value = m1 % m2
        else:
            with pytest.raises(TypeError):
                value = m1 % m2
        if op == "**":
            value = m1 ** m2
        else:
            with pytest.raises(TypeError):
                value = m1 ** m2
        if op == "<<":
            value = m1 << m2
        else:
            with pytest.raises(TypeError):
                value = m1 << m2
        if op == ">>":
            value = m1 >> m2
        else:
            with pytest.raises(TypeError):
                value = m1 >> m2
        if op == "&":
            value = m1 & m2
        else:
            with pytest.raises(TypeError):
                value = m1 & m2
        if op == "^":
            value = m1 ^ m2
        else:
            with pytest.raises(TypeError):
                value = m1 ^ m2
        if op == "|":
            value = m1 | m2
        else:
            with pytest.raises(TypeError):
                value = m1 | m2  # noqa: F841
    assert base.Model._thread_local.operators == {}


def test_model_can_save_to_disk(model_with_no_args):
    temp_file = os.path.join(tempfile.mkdtemp(), "thinc_model")
    model_with_no_args.to_disk(temp_file)


def test_model_can_load_from_disk(model_with_no_args):
    temp_file = os.path.join(tempfile.mkdtemp(), "thinc_model")
    model_with_no_args.to_disk(temp_file)
    m2 = model_with_no_args.from_disk(temp_file)
    assert model_with_no_args.to_bytes() == m2.to_bytes()



# coding: utf8
from __future__ import unicode_literals

from .affine import Affine
from ... import describe
from ...describe import Synapses
from ...check import has_shape
from ... import check


@describe.attributes(
    W=Synapses("Weights matrix", lambda obj: (obj.nO, obj.nI), lambda W, ops: None)
)
class Softmax(Affine):
    name = "softmax"

    @check.arg(1, has_shape(("nB", "nI")))
    def predict(self, input__BI):
        output__BO = self.ops.affine(self.W, self.b, input__BI)
        self.ops.softmax(output__BO, inplace=True)
        return output__BO

    @check.arg(1, has_shape(("nB", "nI")))
    def begin_update(self, input__BI, drop=0.0):
        output__BO = self.predict(input__BI)

        @check.arg(0, has_shape(("nB", "nO")))
        def finish_update(grad__BO, sgd=None):
            self.ops.gemm(grad__BO, input__BI, trans1=True, out=self.d_W)
            self.d_b += grad__BO.sum(axis=0)
            grad__BI = self.ops.gemm(grad__BO, self.W)
            if sgd is not None:
                sgd(self._mem.weights, self._mem.gradient, key=self.id)
            return grad__BI

        return output__BO, finish_update



# coding: utf8
from __future__ import unicode_literals

from srsly import cloudpickle as pickle

from ...neural._classes.feature_extracter import FeatureExtracter


def test_pickle():
    model = FeatureExtracter([100, 200])
    bytes_data = pickle.dumps(model)
    loaded = pickle.loads(bytes_data)
    assert loaded.attrs == model.attrs



# coding: utf8
from __future__ import unicode_literals

import pytest

from ..neural.util import get_ops
from ..neural.ops import NumpyOps, CupyOps


def test_get_ops():
    Ops = get_ops("numpy")
    Ops is NumpyOps
    Ops = get_ops("cpu")
    assert Ops is NumpyOps
    Ops = get_ops("cupy")
    assert Ops is CupyOps
    Ops = get_ops("gpu")
    assert Ops is CupyOps
    with pytest.raises(ValueError):
        Ops = get_ops("blah")



# coding: utf8
from __future__ import unicode_literals

from .model import Model
from ... import describe


def _run_child_hooks(model, X, y):
    for layer in model._layers:
        for hook in layer.on_data_hooks:
            hook(layer, X, y)
        X = layer(X)


@describe.on_data(_run_child_hooks)
class FeedForward(Model):
    """A feed-forward network, that chains multiple Model instances together."""

    name = "feed-forward"

    def __init__(self, layers, **kwargs):
        self._layers = []
        for layer in layers:
            if isinstance(layer, FeedForward):
                self._layers.extend(layer._layers)
            else:
                self._layers.append(layer)
        Model.__init__(self, **kwargs)

    @property
    def input_shape(self):
        return self._layers[0].input_shape

    @property
    def output_shape(self):
        return self._layers[-1].output_shape

    def predict(self, X):
        for layer in self._layers:
            X = layer(X)
        return X

    def begin_update(self, X, drop=0.0):
        callbacks = []
        for layer in self._layers:
            X, inc_layer_grad = layer.begin_update(X, drop=drop)
            callbacks.append(inc_layer_grad)

        def continue_update(gradient, sgd=None):
            for callback in reversed(callbacks):
                if gradient is None or callback is None:
                    break
                gradient = callback(gradient, sgd)
            return gradient

        return X, continue_update



# coding: utf8
from __future__ import unicode_literals

from ...i2v import *  # noqa: F401, F403
from ...v2v import *  # noqa: F401, F403
from ...t2t import *  # noqa: F401, F403
from ...t2v import *  # noqa: F401, F403
from ...api import *  # noqa: F401, F403
from ...misc import *  # noqa: F401, F403
